"""Base package for Cypress Creek Renewables Utilities
"""

__version__ = "0.2.0b10"

from collections import namedtuple
from datetime import (
    date,
    datetime
)
import numpy as np
import pandas as pd

# Custom type definitions
DateLike = datetime|date
ListLike = list|set
SeriesLike = np.ndarray|pd.Series
Numeric = int|float

# Custom data structures
DaylightParams = namedtuple('DaylightParams', ['lat', 'lon', 'tz'])
ProjectModelParams = namedtuple('ProjectModelParams', ['project_name', 'ccr_id', 'folder', 'tz', 'lat', 'lon',
                                             'elevation', 'mwdc', 'racking', 'tilt_gcr', 'max_angle',
                                             'temp_coef', 'a_module', 'b_module', 'delta_tcnd'])

from ccrenew import (
    ccr,
    data_determination
)

# Custom exceptions
class AWSError(Exception):
    pass
class FileNotFoundError(Exception):
    pass
class FileOpenError(Exception):
    pass


print('Pulling df_keys from Smartsheets...')
all_df_keys = ccr.get_df_keys(retired=True)


# -*- coding: utf-8 -*-
import pandas as pd
from ccrenew.dashboard.data_processing.battery_deg import Battery


def Run_AC_Batt(df_kwh, rates, deg_hours):

    # initiate 
    soc = 1065.864
    b = Battery()
    b.power = 499
    b.nameplate_duration = soc/499
    b.limit = 499
    b.charge_eff = 0.88
    b.discharge_eff = 1.0
    b.degradation_hours = deg_hours
    
    # run
    test = b.binary(rates.values, df_kwh.values, soc0 = soc)
    test['SOC'] = test['SOC'][:-1]
    test['inpS'] = df_kwh
    
    df_battery = pd.DataFrame(test, index = df_kwh.index)

    return df_battery



def AC_Batt_PP(df, df_battery):
    POI_limit = 499
    site_limit = 500
    no_load = 0.5 
    copper_loss = site_limit * 0.009
    
    aux = pd.DataFrame([], index = df.index)
    aux['solar+storage'] = df_battery['solar+storage']
    
    #
    aux['MF_XFMR'] =  aux['solar+storage'] - (no_load + (aux['solar+storage']/site_limit)**2 * copper_loss)
    
    #
    AC_wiring_loss = 0.006
    XFMR_year_max = site_limit * 489.646/500.
    aux['AC_wiring'] = aux['MF_XFMR'] * (1-AC_wiring_loss*(aux['MF_XFMR']/XFMR_year_max)**2)
    
    #skipping HV XFMR
    #skipping gen tie
    #skipping PF loss
    #skipping tracker losses
    
    #
    inv_night_loss = 0.570214669
    aux['Inv_night'] = aux['AC_wiring'].copy()
    aux['Inv_night'].loc[aux['Inv_night']<0] = aux['Inv_night'] - inv_night_loss
    
    #
    scada_loss = 1000
    aux['Scada'] = aux['Inv_night'] - (scada_loss*site_limit/8760./1000.)

    ava = 0.99
    aux['POI_cap'] = aux['Scada'].clip(upper=POI_limit)
    aux['ava'] = aux['POI_cap']*ava
    aux['POI'] = aux['ava'].copy()
    aux['POI_no_night'] = aux['POI'].clip(lower=0)

    
    return aux



"""
Package for working with Cypress Creek Renewables Dashboard process
"""
import warnings

from pandas.errors import PerformanceWarning
warnings.simplefilter(action='ignore', category=(FutureWarning, PerformanceWarning)) # type: ignore
warnings.filterwarnings(action='ignore', message='divide by zero')

import boto3
from collections import namedtuple
from datetime import datetime
import os
from pathlib import Path


# Create boto s3 client
s3_client = boto3.client('s3')

# Initilize custom mappings
Colmap = namedtuple('Colmap', ['col_list', 'col_avg'])
DataPlatformDirs = namedtuple('DataPlatformDirs', ['sharepoint', 's3'])
S3FilePath = namedtuple('S3FilePath', ['bucket', 'key'])


def make_s3_filepath(filepath:str|Path, obj_type:str) -> S3FilePath:
    filepath = Path(filepath)
    filepath_parts = filepath.parts
    bucket = filepath_parts[1]
    key = '/'.join(filepath_parts[2:])

    # Add a trailing slash for S3 folders
    if obj_type == 'folder':
        key = key + '/'

    s3_filepath = S3FilePath(bucket, key)

    return s3_filepath

def get_modified_time(filepath:str|Path, data_platform:str, obj_type:str='file'):
    if data_platform == 's3':
        s3_filepath = make_s3_filepath(filepath, obj_type=obj_type)
        s3_file = s3_client.get_object(Bucket=s3_filepath.bucket, Key=s3_filepath.key)
        modified_time = s3_file['LastModified']
    else:
        modified_time = datetime.fromtimestamp(os.path.getmtime(filepath))

    return modified_time


print('Importing DashboardSession')
from ccrenew import (
    all_df_keys,
    cloud_data
)
from ccrenew.dashboard.utils.dashboard_utils import func_timer
from ccrenew.data_determination import daylight
from ccrenew.dashboard.plotting.plots import Plotter
from ccrenew.dashboard.session import DashboardSession
from ccrenew.dashboard.project import Project
from ccrenew.dashboard.utils.project_neighbors import (
    find_nearby_projects,
    find_nearby_similar_projects,
    find_similar_projects
)


# Helper functions
class Helpers:
    def __init__(self):
        self.df_keys = all_df_keys
        self.find_nearby_projects = find_nearby_projects
        self.find_nearby_similar_projects = find_nearby_similar_projects
        self.find_similar_projects = find_similar_projects
        self.get_sat_weather = cloud_data.get_sat_weather
        self.daylight = daylight

helpers = Helpers()



# -*- coding: utf-8 -*-
"""
Module to coordinate processing of CCRenew Dashboard projects.
"""
from __future__ import annotations

import boto3
import builtins
from ccrenew.dashboard.utils.logging_conf import get_logging_config
from datetime import datetime
from dateutil import parser
import logging
import logging.config
from matplotlib.pyplot import waitforbuttonpress
import os
import pandas as pd
from pandas.io.sql import SQLTable
from pathlib import Path
import pickle
import s3fs
import sys
import time
import traceback
import warnings

from ccrenew import (
    __version__,
    ccr,
    all_df_keys
)
from ccrenew.dashboard import (
    DataPlatformDirs,
    func_timer,
    get_modified_time,
    make_s3_filepath
)
from ccrenew.dashboard import Plotter
from ccrenew.dashboard.project import Project
from ccrenew.dashboard.utils.df_tools import df_update_join
import ccrenew.dashboard.utils.dashboard_utils as utils
import ccrenew.dashboard.utils.project_neighbors as neighbs

from ccrenew.dashboard.data_processing.batch_process import (
    export_pool,
    process_pool)
    

# Create boto s3 client
s3_client = boto3.client('s3')

# suppress warnings about "A value is trying to be set on a copy of a slice from a DataFrame."
pd.set_option('mode.chained_assignment', None)

# Initialize logger
# Get logging configuration from the logging.conf file in the same directory as this file
username = os.getenv('username')
dashboard_folder = ccr.file_project
# log_filename = '{}_dashboard-{}.log'.format(datetime.now().strftime('%Y-%m-%d_%H-%M-%S'), username)
# log_config = os.path.join(dashboard_folder, 'Python_Functions', 'logging.conf')
# log_file = os.path.join(dashboard_folder, 'Python_Functions', '_old', 'logs', log_filename)
# LOGGING_CONFIG = get_logging_config(log_file)
# logging.config.dictConfig(LOGGING_CONFIG)

# Create logger
# logger = logging.getLogger(__name__)
# logger.info('Beginning Dashboard Session')

def _execute_insert(self, conn, keys, data_iter):
    #print "Using monkey-patched _execute_insert"
    data = [dict((k, v) for k, v in zip(keys, row)) for row in data_iter]

    # Python 2 compatibility:
    try:
        conn.execute(self.insert_statement().values(data))
    except AttributeError:
        conn.execute(self.table.insert().values(data))
    

SQLTable._execute_insert = _execute_insert # type: ignore

    
class DashboardSession:
    """Main orchestrator for processing, plotting, and exporting data for projects (i.e. sites).

    Args:
        project_list (str or list): List of [Projects][ccrenew.dashboard.project.Project] to initialize when initializing the session.
        data_cutoff_date (str or datetime, optional): The last date you want to analyze. Defaults to the current date the dashboard is being run.
        data_source (str, optional): The location where project data is stored.
            Default is current year Sharepoint data. To pull S3 data add 's3' to the variable.
        
        `data_source` examples:
        
        | `data_source`  | Explanation                  |
        |----------------|------------------------------|
        | `None`         | Current year Sharepoint data |
        | `s3`           | Current year S3 data         |
        | `2022_data_`   | 2022 Sharepoint data         |
        | `2022_data_s3` | 2022 S3 data                 |
    """

    # Instantiate globals as class variables
    df_keys = all_df_keys.copy()
    __version__ = __version__

    print('Pulling diagnostic comments from Smartsheets...')
    # Smartsheet comments from previous reporting months. I'm going to hardcode Jan 2021 as the earliest date (column index 69 - Nice) but we can change that easily
    df_ss_comments = ccr.get_ss_as_df('2819883898562436', drop_col_1=False, start_col=69, index_col='Project')

    # Site lists
    tracker_sites = ['5840 Buffalo Road', 'Alexis', 'ATOOD II', 'Bar D', 'Barnhill Road Solar', 'Bay Branch',
                     'Bonanza', 'Bovine Solar', 'Bronson', 'Cascade', 'Chisum', 'Copperfield', 'Curie',
                     'Eddy II', 'Gaston II', 'Griffin', 'Grove Solar (NC)', 'Grove', 'Hardison Farm',
                     'Hopewell Friends', 'Hyline', 'IS - 46 Tracker', 'IS 67', 'Lampwick', 'Leggett Solar',
                     'Mars', 'Neff', 'Nimitz', 'Open Range', 'Palmetto Plains', 'Prince George 1',
                     'Prince George 2', 'Railroad', 'Shoe Creek', 'Siler', 'Simba', 'Springfield', 'Sterling',
                     'Thunderegg', 'Vale', 'Wagyu', 'Warren', 'Wendell', 'West Moore II', 'Yellow Jacket']
    battery_ac_funds = ['LGE']
    battery_ac_sites = df_keys.query("Fund in @battery_ac_funds").index.tolist()
    battery_dc_sites = ['Salt Point', 'Dubois', 'Landau II', 'Clendenin A']
    
    # Type annotations & docstrings for documentation
    df_keys: pd.DataFrame
    """Metadata for all projects."""
    df_ss_comments: pd.DataFrame
    """DataFrame with Smartsheet comments."""
    battery_ac_funds: list
    """Funds with projects with AC batteries."""
    battery_ac_sites: list
    """Projects with AC batteries."""
    battery_dc_sites: list
    """Projects with DC batteries."""
    tracker_sites: list
    """Projects with tracker-type racking."""

    def __init__(self,
                 project_list=None,
                 data_cutoff_date=None,
                 data_source=None,
                 **kwargs):

        print('Initializing DashboardSession...')
        
        # Initialize a dict to store projects - key = project name; value = project object
        # utils.project_dict() overrides the __repr__ method of dict to pretty print the projects when you call it
        self.project_list = utils.project_dict()

        # Add projects to the session if provided
        if project_list:
            if isinstance(project_list, (str, list)):
                self.add_projects(project_list)
            else:
                print('`project_list` must be a string or list. The projects provided were not added to the DashboardSession. Please try again in the initialized DashboardSession.')

        # Parse & add data_cutoff_date if provided. Default to current date if not provided or error during parsing 
        if data_cutoff_date:
            try:
                self.data_cutoff_date = parser.parse(data_cutoff_date)
            except:
                self.data_cutoff_date = datetime.today()
        else:
            self.data_cutoff_date = datetime.today()

        def set_data_source(data_source:str|None) -> None:
            if data_source:
                self.data_source = data_source
            else:
                self.data_source = f"{datetime.now().year}_data_"

        self.data_platform_dirs = DataPlatformDirs(ccr.file_project,
                                                   's3://cypress-perfeng-datalake-dev-us-west-2/Raw/HourlyResolution/')

        set_data_source(data_source)
        if 's3' in self.data_source.lower():
            data_source = self.data_source.replace(' ', '').replace('s3', '')
            set_data_source(data_source)
            self.dashboard_dir = self.data_platform_dirs.s3
            self.data_platform = 's3'
        else:
            self.dashboard_dir = self.data_platform_dirs.sharepoint
            self.data_platform = 'sharepoint'

        # Get snow dataframe
        self.raw_snow_df, self.snow_file = utils.get_snow_df(dashboard_folder, self.data_source, file_format='csv', data_platform=self.data_platform)

        # Create list for projects that have errored out
        self.errored_projects = {}

        # List of plotter objects for different projects
        self.plotters = {}

        # Type annotations & docstrings for documentation
        self.project_list: dict
        """List of [Projects][ccrenew.dashboard.project.Project] that
        have been initialized & added to the
        [DashboardSession][ccrenew.dashboard.session.DashboardSession] instance."""
        self.data_platform: str
        """Backend architecture of the source data for a project. Either `s3` or `sharepoint`."""
        self.data_source: str
        """Location of the source data for projects based on production year."""
        self.raw_snow_df: pd.DataFrame
        """Snow data."""
        self.data_cutoff_date: datetime
        """Last day to process data."""
        self.errored_projects: dict
        """List of [Projects][ccrenew.dashboard.project.Project] that have encountered
        errors while processing & their error messages."""
        self.plotters: dict
        """List of projects that have been plotted & their plots."""

    def __repr__(self):
        return 'DashboardSession object with {} projects'.format(len(self.project_list))


    def __str__(self):
        return 'DashboardSession object with {} projects'.format(len(self.project_list))


    def get_project(self, project_name:str|Project, get_neighbors:bool=True, report_type:str = 'monthly') -> Project:
        """
        Return a [Project][ccrenew.dashboard.project.Project] object given a project name.

        Args:
            project_name (str or Project): Name of a [Project][ccrenew.dashboard.project.Project] to return from the
                [DashboardSession][ccrenew.dashboard.session.DashboardSession] instance.
                If project does not yet exist in the session, it will be added then returned.
            get_neighbors (bool, optional): Flag to add & initialize neighbor [Projects][ccrenew.dashboard.project.Project].
            report_type (str): Method for processing the project. Options are `monthly`, `ul`, or `weekly`.

        Returns:
            Project object for the given the project name.
                If the project does not exist in the session when the call is made, the project will be initialized & returned.
        """
        if isinstance(project_name, Project):
            project = project_name
        else:
            try:
                project = self.project_list[project_name]
            except KeyError:
                self.add_project(project_name, get_neighbors=get_neighbors, report_type=report_type)
                project = self.project_list[project_name]
 
        return project


    def add_project(self, project_name:str|Project|list, get_neighbors:bool = True, report_type:str = 'monthly') -> Project:
        """Adds & initializes a [Project][ccrenew.dashboard.project.Project]
        to a [DashboardSession][ccrenew.dashboard.session.DashboardSession] instance.

        Args:
            project_name (str or Project): The name of the project to add to the [DashboardSession][ccrenew.dashboard.session.DashboardSession] instance.
            get_neighbors (bool, optional): Flag to add & initialize neighbor [Projects][ccrenew.dashboard.project.Project].
            report_type (str): Method for processing the project. Options are `monthly`, `ul`, or `weekly`.
        """
        # Type validation
        if not isinstance(project_name, (str, list, Project)):
            raise TypeError("`project_name` must be a string, Project object, or Python list! Please reformat input and try again.")

        # Determine behavior based on what was passed to method
        if isinstance(project_name, list):
            print(f"A list was supplied to the `add_project` method. The method will return the first project in the list that was successfully added to the `DashboardSession`.")
            project_list = project_name
            project = self.add_projects(project_list, get_neighbors=get_neighbors, report_type=report_type)
        elif project_name in self.project_list:
            project = self.project_list[project_name]
        elif isinstance(project_name, Project):
            project = project_name
            self.project_list[project.project_name] = project
        else:
            project = self.__initialize_project(project_name, get_neighbors=get_neighbors, report_type=report_type)
            if not project.error_info:
                self.project_list[project.project_name] = project

        return project
    

    def add_projects(self, project_list:list|str|Project, get_neighbors:bool = True, report_type:str = 'monthly') -> Project:
        """Calls [add_project()][ccrenew.dashboard.session.DashboardSession.add_project] for each project in a list.

        Args:
            project_list (list, str): list of project names to add & initialize to the [DashboardSession][ccrenew.dashboard.session.DashboardSession] instance.
            get_neighbors (bool, optional): See [add_project()][ccrenew.dashboard.session.DashboardSession.add_project].
            report_type (str): Method for processing the project. Options are `monthly`, `ul`, or `weekly`.
        """
        if isinstance(project_list, (str, Project)):
            project = self.add_project(project_list, get_neighbors=get_neighbors, report_type=report_type)
            return project
        elif isinstance(project_list, list):
            successful_projects = []
            for project_name in project_list:
                try:
                    project = self.add_project(project_name, get_neighbors=get_neighbors, report_type=report_type)
                    successful_projects.append(project)
                except Exception as e:
                    print(f"Could not add {project_name} to the `DashboardSession` instance. See error below:")
                    print(e)
                    continue
            
            # Return the first successful project from the list
            project = successful_projects[0]
            return project
        else:
            raise TypeError("`project_list` must be a string, Project object, or Python list! Please reformat input and try again.")
                

    def __initialize_project(self, project_name:str, get_neighbors:bool = True, report_type:str = 'monthly'):
        """Initializes a project object with metadata

        Args:
            project_name (str): Name of a project to initialize.
            get_neighbors (bool): Flag to add & initialize neighbor projects.
            report_type (str): Method for processing the project. Options are `monthly`, `ul`, or `weekly`.

        Returns:
            Project: An initialized Project object
        """

        try:
            df_proj_keys = self.df_keys.query("Project == @project_name").to_dict('records')[0]
            df_proj_ss_comments = self.df_ss_comments.query("index == @project_name").to_dict('records')[0]
        except IndexError:
            # Log that project was not found
            warn_msg = '"{}" not found in df keys. Please verify the project is present in df keys & the spelling is correct and try again'.format(project_name)
            print(warn_msg)
            # logger.warn(warn_msg)
            df_proj_keys = {'Racking': 'None'}
            df_proj_ss_comments = {}
            proj_init_dict = {'error_info': warn_msg}

        # Get neighbor projects
        if get_neighbors:
            neighbor_list = self.__get_neighbors(project_name)
        else:
            neighbor_list = {}

        # Set battery/tracker attributes
        Battery_AC_site = project_name in self.battery_ac_sites
        Battery_DC_site = project_name in self.battery_dc_sites
        Tracker_site = df_proj_keys['Racking'] == 'Tracker'
        
        proj_init_dict = {}
        proj_init_dict['project_name'] = project_name
        proj_init_dict['df_proj_keys'] = df_proj_keys
        proj_init_dict['df_proj_ss_comments'] = df_proj_ss_comments
        proj_init_dict['dashboard_dir'] = self.dashboard_dir
        proj_init_dict['data_cutoff_date'] = self.data_cutoff_date
        proj_init_dict['data_platform'] = self.data_platform
        proj_init_dict['data_platform_dirs'] = self.data_platform_dirs
        proj_init_dict['data_source'] = self.data_source
        proj_init_dict['report_type'] = report_type
        proj_init_dict['Battery_AC_site'] = Battery_AC_site
        proj_init_dict['Battery_DC_site'] = Battery_DC_site
        proj_init_dict['Tracker_site'] = Tracker_site
        proj_init_dict['raw_snow_df'] = self.raw_snow_df
        proj_init_dict['snow_file'] = self.snow_file
        proj_init_dict['snow_file'] = self.snow_file
        proj_init_dict['neighbor_list'] = neighbor_list

        try:
            project = Project(proj_init_dict)
        except Exception as e:
            error_info = sys.exc_info()
            error_type = error_info[:2]
            traceback = error_info[2]
            error_source = traceback.tb_next.tb_frame.f_code # type: ignore
            lineno = traceback.tb_next.tb_lineno # type: ignore
            file = error_source.co_filename
            func = error_source.co_name
            
            print(f"Error initializing {project_name}.")
            print(f"Error details: {error_type} on line {lineno} of `{func}` in {file}")
            proj_init_dict['error_info'] = f"{error_type} on line {lineno} of `{func}` in {file}"
            project = Project(proj_init_dict)

        return project
    

    def __get_neighbors(self, project_name:str):
        # Find all neighbors that satisfy the distance and equipment requirements
        all_neighbors = neighbs.find_nearby_projects(project_name, print_data=False, include_retired=False).index.tolist()
        poa_neighbors = neighbs.find_nearby_similar_projects(project_name, print_data=False, include_retired=False).index.tolist()
        
        # Then remove the search project from the list
        # (Sometimes it will return an empty list if there is some data missing in `df_keys`) so we need to check if it returned a non-empty list first
        try:
            all_neighbors.remove(project_name)
        except ValueError:
            pass
        try:
            poa_neighbors.remove(project_name)
        except ValueError:
            pass

        # We'll create a neighbor_list dictionary
        # Value=True: we can use it for POA data sub
        # Value=False: we can only use it for weather data sub
        neighbor_list = {project_name: True if project_name in poa_neighbors else False for project_name in all_neighbors}

        # If the project has neighbors, we'll add them to the DashboardSession instance
        # Set get_neighbors to False so we don't get neighbors of neighbors of neighbors etc.
        if neighbor_list:
            for neighbor in neighbor_list.copy():
                try:
                    self.add_project(neighbor, get_neighbors=False)
                except:
                    neighbor_list.pop(neighbor)
        
        return neighbor_list


    def process_project(self, project_name:str|Project|list, reprocess:bool = False, datasub:bool = False,
                        use_solcast:bool = True, report_type:str = 'monthly', **kwargs) -> Project:
        """
        Process data for the given [Project][ccrenew.dashboard.project.Project].

        Args:
            project_name (str or Project): The name of the [Project][ccrenew.dashboard.project.Project] to process.
            reprocess (bool): Whether to reprocess the data. This option will likely
                be uncommon as any changes to a [Project's][ccrenew.dashboard.project.Project] config file or powertrack file will
                automatically reprocess that data. This could be used if any changes are made
                outside of the config/powertrack file or a change is made to a neighbor.
            datasub (bool): Option to use automatic data substitution algorithm.
            use_solcast (bool): Option to use Solcast data in data sub process. `datasub`
                must be set to `True` for this option to take effect.
            report_type (str): Method for processing the project. Options are `monthly`, `ul`, or `weekly`.
        Returns:
            Processed Project object.
        """
        if isinstance(project_name, list):
            # If we call this method with a list we'll default to series processing & return the first project in the list
            project_list = project_name
            self.process_projects(project_list=project_list, method='series', reprocess=reprocess,
                                  datasub=datasub, use_solcast=use_solcast, report_type=report_type, **kwargs)

            project = self.project_list[project_list[0]]
            
            return project

        elif isinstance(project_name, (str, Project)):
            start = time.time()

            # Get project object
            project = self.get_project(project_name, report_type=report_type)
            
            # If the project has already been added to the session but we want to run it with a different process method
            # We'll update the process method & force the config file to reload
            if project.report_type != report_type:
                project.report_type = report_type
                project.last_update_config = datetime.fromtimestamp(0)

            # Pull the last updated dates from the filesystem for the config, bool, & Powertrack files
            try:
                last_update_config_file = get_modified_time(project.config_filepath, self.data_platform, 'file')
                last_update_bool_file = get_modified_time(project.bool_filepath, self.data_platform, 'file')
                last_update_powertrack_file = get_modified_time(project.powertrack_filepath, self.data_platform, 'file')
            except:
                self.__update_project_filepaths(project)
                last_update_config_file = get_modified_time(project.config_filepath, self.data_platform, 'file')
                last_update_bool_file = get_modified_time(project.bool_filepath, self.data_platform, 'file')
                last_update_powertrack_file = get_modified_time(project.powertrack_filepath, self.data_platform, 'file')

            # Check if the config, bool, or Powertrack files have been updated.
            # If not we don't need to process.
            if reprocess:
                project.last_update_config = datetime.fromtimestamp(0)
                project.last_update_bool = datetime.fromtimestamp(0)
                project.last_update_powertrack = datetime.fromtimestamp(0)
            if any([project.last_update_config != last_update_config_file,
                    project.last_update_bool != last_update_bool_file,
                    project.last_update_powertrack != last_update_powertrack_file]):

                # Pull config and/or Powertrack data
                self.__prepare_source_data(project)
            else:
                if project.processed:
                    print('{} already processed'.format(project.project_name))

                    return project

            # Process data - update neighbor sensors first then process the project
            print(f'Processing {project_name}....')
            # project.neighbor_list = self.__get_neighbors(project_name)
            self.__get_neighbor_sensor_data(project)
            failed_neighbors = []
            for neighbor_name in project.neighbor_list:
                try:
                    neighbor = self.project_list[neighbor_name]
                    self.__prepare_source_data(neighbor)
                    neighbor._calculate_native_sensor_averages()
                except Exception as e:
                    print(f"Neighbor for {project.project_name} failed while processing")
                    print(f"Neighbor name: {neighbor_name}")
                    print(f"Error message: {e}")
                    failed_neighbors.append(neighbor_name)
                    continue
            for neighbor_name in failed_neighbors:
                del project.neighbor_list[neighbor_name]
            neighbor_sensor_data = {self.project_list[neighbor].project_name: self.project_list[neighbor].df_sensors_native_avg for neighbor in project.neighbor_list}

            try:
                project._process_data(neighbor_sensor_data=neighbor_sensor_data,
                                      datasub=datasub, use_solcast=use_solcast)
            except Exception:
                print('\n')
                print('###########################')
                print(f'{project.project_name} not processed. See below for error information.')
                print(traceback.format_exc())
                print('###########################')

                project.last_update_powertrack = datetime.fromtimestamp(0)
                project.errored = True
                project.error_info = traceback.format_exc()

                return project
            project.processed = True
            project.errored = False
            print('{} successfully processed. Processing time: {:.2f}s'.format(project.project_name, time.time()-start))

            return project
        else:
            raise TypeError("`project_name` must be a string or Project object! Please reformat input and try again.")
    
    
    def process_projects(self, project_list:str|Project|list, method:str = 'series', reprocess:bool = False,
                         datasub:bool = False, use_solcast:bool = True, report_type:str = 'monthly', **kwargs):
        """Calls [process_project()][ccrenew.dashboard.session.DashboardSession.process_project] for each
        [Project][ccrenew.dashboard.project.Project] in a list.

        Args:
            project_list (list): List of [Project][ccrenew.dashboard.project.Project] names process.
            method (str): Option to perform processing operations in series or parallel.
            reprocess (bool): See [process_project()][ccrenew.dashboard.session.DashboardSession.process_project].
            method (str): Option to perform processing operations in series or parallel.
            reprocess (bool): See [process_project()][ccrenew.dashboard.session.DashboardSession.process_project].
            datasub (bool): Option to use automatic data substitution algorithm.
            use_solcast (bool): Option to use Solcast data in data sub process. `datasub`
                must be set to `True` for this option to take effect.
            report_type (str): Method for processing the project. Options are `monthly`, `ul`, or `weekly`.
        """
        if isinstance(project_list, (str, Project)):
            project_name = project_list
            project = self.process_project(project_name, reprocess=reprocess, datasub=datasub, use_solcast=use_solcast)

            return project
        elif isinstance(project_list, list):
            if len(project_list) == 1:
                project_name = project_list[0]
                project = self.process_project(project_name, reprocess=reprocess, datasub=datasub, use_solcast=use_solcast)

                return project
            else:
                start = time.time()
                if method == 'series':
                    for i, project_name in enumerate(project_list):
                        project = self.process_project(project_name, reprocess=reprocess, datasub=datasub, use_solcast=use_solcast)
                        if not project:
                            continue

                        elapsed = time.time() - start
                        print('*******************************************************')
                        print(f'{project_name} complete. {i+1} of {len(project_list)} projects completed. Total elapsed time: {elapsed:2f}s')
                        print('*******************************************************')

                elif __name__ == 'ccrenew.dashboard.session':
                    # We'll use half of the available processors for the machine to not completely bog it down.
                    max_workers = int(os.cpu_count()/2) # type: ignore
                    print(f"Processing {len(project_list)} projects with {min(max_workers, len(project_list))} workers.")
                    print("Progress messages from workers will be suppressed in an interactive session.")
                    print("Projects will report success upon completion and may not complete in the order in which they were submitted.")

                    # Construct a dict of kwargs for each project to pass to `process_pool`
                    if self.data_platform == 's3':
                        data_source = self.data_source + 's3'
                    else:
                        data_source = self.data_source

                    kwargslist =  [{"project_name": project_name,
                                    "session": DashboardSession,
                                    "data_source": data_source,
                                    "reprocess": reprocess,
                                    "datasub": datasub,
                                    "use_solcast": use_solcast,
                                    "project_num": i+1,
                                    "project_total": len(project_list),
                                    "batch_start": start
                                    } for i, project_name in enumerate(project_list)]

                    # Results are a list of Future objects, where we can get the results of the function call from `.result()`
                    results = process_pool(kwargslist)
                    for result in results:
                        processed_project_list = result.result()['project_list']
                        for project_dict in processed_project_list:
                            try:
                                project_name, project = project_dict
                                # If the project is already processed in the parent session we won't add it
                                if project_name in self.project_list and self.project_list[project_name].processed:
                                    pass
                                else:
                                    self.project_list[project_name] = project
                            except:
                                continue

                elapsed = time.time() - start
                print(f"\n{len(project_list)} projects processed. Elapsed time: {elapsed:.2f}s.\n")
        else:
            raise TypeError("`project_list` must be a string or Python list! Please reformat input and try again.")


    def process_all_projects(self, method:str = 'parallel', reprocess:bool = True, datasub:bool = False,
                             use_solcast:bool = True, report_type:str = 'monthly', **kwargs):
        """
        Processes all projects in the [project_list][ccrenew.dashboard.session.DashboardSession.project_list]
        for the [DashboardSession][ccrenew.dashboard.session.DashboardSession] instance.

        Args:
            method (str): Option to perform operations in series or parallel.
            reprocess (bool): See [process_project()][ccrenew.dashboard.session.DashboardSession.process_project].
            datasub (bool): Option to use automatic data substitution algorithm.
            use_solcast (bool): Option to use Solcast data in data sub process. `datasub`
                must be set to `True` for this option to take effect.
            report_type (str): Method for processing the project. Options are `monthly`, `ul`, or `weekly`.
        """
        project_list = list(self.project_list.keys())
        self.process_projects(project_list, method=method, reprocess=reprocess,
                              datasub=datasub, use_solcast=use_solcast, report_type=report_type)
    

    def export_project(self, project_name:str|list, server:str = 'dev', dest:str|list = 'all',
                       report_type:str = 'monthly', use_solcast:bool = True, save_pickle:bool = True, **kwargs) -> None:
        """
        Export project to desired location(s).

        Args:
            project_name (str or Project): Name of a project to export.
            server (str): Environment . Options are `prod` or `dev`, or `datasub`.
            dest (str or list): Option to export only certain parts of the project. `all` will export to all of the below destinations.
            report_type (str): Option to select which report type to run. Options are `monthly`, `ul`, and `weekly`.
            save_pickle (bool): Option to save pickle after exporting.
            **kwargs: Optional arguments passed to [process_project()][ccrenew.dashboard.project.Project.process_project].

        `dest` Options

        | dest      | Description                                                              |
        |-----------|--------------------------------------------------------------------------|
        |  `excel`  | Export dataframes to Excel dashboard template.                          |
        |   `df`    | Export dataframe to the `Dataframes` folder of the project's [project_directory][ccrenew.dashboard.project.Project.project_directory], S3, and Bartertown PostgreSQL database. |
        | `summary` | Export monthly summary data to S3 and Bartertown PostgreSQL database.   |
        |  `snow`   | Export [snow_data][ccrenew.dashboard.project.Project.snow_data] and [snow_coverage][ccrenew.dashboard.project.Project.snow_coverage] dataframes to S3. |
        |  `bool`   | Export bool file to the `Plant_Config_File` directory of the project's [project_directory][ccrenew.dashboard.project.Project.project_directory]. |

        `server` SQL Destinations

        |   server   | `df`                  | `summary`               |
        |------------|-----------------------|-------------------------|
        |   `prod`   | am_processed_data     | am_summary_data         |
        |   `dev`    | am_processed_data_dev | am_summary_data         |
        | `datasub` | am_processed_data_dev | am_summary_data_datasub |

        `server` S3 Destinations

        |   server   | bucket                                         | `df` prefix          | `summary` prefix      |
        |------------|------------------------------------------------|----------------------|-----------------------|
        |   `prod`   | s3://cypress-perfeng-datalake-dev-us-west-2    | Curated/HourlyLosses | Curated/MonthlyLosses |
        |   `dev`    | s3://cypress-perfeng-datalake-onprem-us-west-2 | Curated/HourlyLosses | Curated/MonthlyLosses |
        | `datasub`  | s3://cypress-perfeng-datalake-onprem-us-west-2 | Curated/HourlyLosses | Curated/MonthlyLosses |
        """
        # Pass function arguments to `export_projects()`.
        # Remove `self`, i.e. the DashboardSession instance
        func_args = locals().copy()
        func_args.pop('self')

        if isinstance(project_name, list):
            # Rename project_name to project_list for `export_projects`
            func_args.pop('project_name')
            func_args['project_list'] = project_name
            self.export_projects(**func_args)
        elif isinstance(project_name, (str, Project)):
            project = self.get_project(project_name, report_type=report_type)

            # If we're only exporting the bool file then we don't need to process
            if dest != 'bool':
                func_args['save_pickle'] = False
                datasub = False
                if server == 'datasub':
                    datasub = True
                self.process_project(project_name, datasub=datasub, use_solcast=use_solcast,
                                     report_type=report_type, **kwargs)

            func_args.pop('project_name')
            project.export(**func_args)
            print(f"{project.project_name} successfully exported.")
        else:
            raise TypeError("`project_name` must be a string or Project object! Please reformat input and try again.")

  
    def export_projects(self, project_list:list|str, server:str = 'dev', dest:str|list = 'all', report_type:str = 'monthly',
                        method:str = 'series', use_solcast:bool = True, save_pickle:bool = True, **kwargs) -> None:
        """Calls [export_project()][ccrenew.dashboard.session.DashboardSession.export_project] for each
        [Project][ccrenew.dashboard.project.Project] in a list.

        Args:
            project_list (list): List of [Project][ccrenew.dashboard.project.Project] names process.
            server (str): See [export_project()][ccrenew.dashboard.session.DashboardSession.export_project].
            dest (str or list): See [export_project()][ccrenew.dashboard.session.DashboardSession.export_project].
            method(str): Option to export in series or parallel.
            save_pickle (bool): See [export_project()][ccrenew.dashboard.session.DashboardSession.export_project].
        """
        # Store function arguments in case we need to pass them to `export_project()`.
        # Argument 0 is `self` so we'll pass arguments 1 and on
        func_args = locals().copy()
        func_args.pop('self')
        method = func_args.pop('method')

        if isinstance(project_list, (str, Project)):
            # Rename project_list to project_name for `export_project`
            func_args.pop('project_list')
            func_args['project_name'] = project_list
            self.export_project(**func_args)
        elif isinstance(project_list, list):
            # If a list was passed, we'll pop the project list & use the remaining args
            project_list = func_args.pop('project_list')

            if len(project_list) == 1:
                project_name = project_list[0]
                self.export_project(project_name, **func_args)
            else:
                start = time.time()
                if method == 'series':
                    for i, project_name in enumerate(project_list):
                        self.export_project(project_name, **func_args)
                    
                        elapsed = time.time() - start
                        print('*******************************************************')
                        print(f'{project_name} complete. {i+1} of {len(project_list)} projects exported. Total elapsed time: {elapsed:2f}s')
                        print('*******************************************************')
                
                elif __name__ == 'ccrenew.dashboard.session':

                    # We'll use half of the available processors for the machine to not completely bog it down.
                    max_workers = int(os.cpu_count()/2) # type: ignore
                    print(f"Exporting {len(project_list)} projects with {min(max_workers, len(project_list))} workers.")
                    print("Progress messages from workers will be suppressed in an interactive session.")
                    print("Projects will report success upon completion and may not complete in the order in which they were submitted.")

                    # Construct a dict of kwargs for each project to pass to `process_pool`
                    if self.data_platform == 's3':
                        data_source = self.data_source + 's3'
                    else:
                        data_source = self.data_source

                    kwargslist =  [{"project_name": project_name,
                                    "session": DashboardSession,
                                    "data_source": data_source,
                                    "func_args": func_args,
                                    "project_num": i+1,
                                    "project_total": len(project_list),
                                    "batch_start": start
                                    } for i, project_name in enumerate(project_list)]

                    # Results are a list of Future objects, where we can get the results of the function call from `.result()`
                    results = export_pool(kwargslist)
                    for result in results:
                        exported_project_list = result.result()['project_list']
                        for project_dict in exported_project_list:
                            try:
                                project_name, project = project_dict
                                # If the project is already processed in the parent session we won't add it
                                if project_name in self.project_list and self.project_list[project_name].processed:
                                    pass
                                else:
                                    self.project_list[project_name] = project
                            except:
                                continue

                elapsed = time.time() - start
                print(f"\n{len(project_list)} projects exported. Elapsed time: {elapsed:.2f}s.\n")                        

        else:
            raise TypeError("`project_list` must be a string or Python list! Please reformat input and try again.")


    def remove_project(self, project_name:str) -> None:
        """Remove the selected project from the active session.

        Args:
            project_name (str): The name of the [Project][ccrenew.dashboard.project.Project]
                to remove from the [DashboardSession][ccrenew.dashboard.session.DashboardSession] instance.
        """
        del self.project_list[project_name]

    def get_ss_comments(self, project_name:str, num_comments: int|None=None) -> None:
        """Prints Smartsheet comments for the [Project][ccrenew.dashboard.project.Project].

        Args:
            project_name (str): Name of the [Project][ccrenew.dashboard.project.Project] to pull comments for.
            num_comments (int, optional): Number of comments to show, starting from the
                most recent. I.e. `5` will show the most recent 5 comments in the
                Smartsheet. Defaults to None, which will show all comments starting in Jan 2021.
        """
        project = self.get_project(project_name, get_neighbors=False)
        project.get_ss_comments(num_comments=num_comments)


    def save_pickle(self, project_name, store_plots=False):
        """Saves the [Project][ccrenew.dashboard.project.Project] to a serialized pickle in its
        [project_directory][ccrenew.dashboard.project.Project.project_directory].

        Args:
            project_name (str or Project): The name of the [Project][ccrenew.dashboard.project.Project] to pickle.
            store_plots (bool, optional): Option to store any plots that have been drawn with the project to the pickle.

        Raises:
            TypeError: if the wrong type is supplied to `project_name`.
        """
        if isinstance(project_name, list) and not isinstance(project_name, str):
            self.save_pickles(project_name, store_plots=store_plots)
        elif isinstance(project_name, (str, Project)):
            project = self.process_project(project_name)
            project.save_pickle(store_plots=store_plots)
        else:
            raise TypeError("`project_name` must be a string or Project object! Please reformat input and try again.")


    def save_pickles(self, project_list, store_plots=False):
        """Calls [pickle_project()][ccrenew.dashboard.session.DashboardSession.pickle_project]
        for each [Project][ccrenew.dashboard.project.Project] in a list.

        Args:
            project_list (list, str, or Project): List of [Project][ccrenew.dashboard.project.Project] names to pickle.
            store_plots (bool, optional): See [pickle_project()][ccrenew.dashboard.session.DashboardSession.pickle_project].
        """
        if isinstance(project_list, (str, Project)):
            project_name = project_list
            self.save_pickle(project_name, store_plots=store_plots)
        elif isinstance(project_list, list):
            for i, project_name in enumerate(project_list):
                print('*******************************************************')
                print(f'Pickling {project_name} - Project {i+1} of {len(project_list)}.')
                print('*******************************************************')
                self.save_pickle(project_name, store_plots=store_plots)
        else:
            raise TypeError("`project_list` must be a string or Python list! Please reformat input and try again.")


    def pickle_project(self, project_name, store_plots=False):
        """Alias for [save_pickle][ccrenew.dashboard.session.DashboardSession.save_pickle].
        
        Raises:
            FutureWarning: This method will be deprecated in a future version. Use [save_pickle][ccrenew.dashboard.session.DashboardSession.save_pickle] instead.
        """

        warnings.warn('pickle_project will be deprecated in a future version. Please use `save_pickle` instead', FutureWarning)
        self.save_pickle(project_name, store_plots=store_plots)


    def pickle_projects(self, project_list, store_plots=False):
        """Alias for [save_pickles][ccrenew.dashboard.session.DashboardSession.save_pickles].
        Raises:
            FutureWarning: This method will be deprecated in a future version. Use [save_pickle][ccrenew.dashboard.session.DashboardSession.save_pickle] instead.
        warnings.warn('pickle_project will be deprecated in a future version. Please use `save_pickles` instead', FutureWarning)
        """
        self.save_pickles(project_list, store_plots=store_plots)


    def load_pickle(self, project_name, year=None, data_cutoff_date=None, show_plots=False):
        """Loads a pickled [Project][ccrenew.dashboard.project.Project] from file.

        Args:
            project_name (str): The name of the [Project][ccrenew.dashboard.project.Project] to load.
            year (int, optional): Deprecated. Year for the pickle will always load the year to match the session.
            data_cutoff_date (str or datetime): Cutoff date for the [Project][ccrenew.dashboard.project.Project]. When
                loading a project that was pickled some time ago the [data_cutoff_date][ccrenew.dashboard.session.DashboardSession.data_cutoff_date]
                could prevent the data from loading to the current date. Default is None,
                which will reset the [data_cutoff_date][ccrenew.dashboard.session.DashboardSession.data_cutoff_date] to the current day.
            show_plots (bool, optional): Option to show plots if pickle was stored with plots saved.
        """
        year = self.data_source.split('_')[0]
        try:
            project_directory = self.df_keys.query("Project == @project_name")['Folder'].item()
        except IndexError:
            # Log that project was not found
            warn_msg = '"{}" not found in df keys. Please verify the project is present in df keys & the spelling is correct and try again'.format(project_name)
            print(warn_msg)
            # logger.warn(warn_msg)
            return

        pickle_jar = os.path.join(self.dashboard_dir,
                                  project_directory,
                                  project_name,
                                  'Pickle_Jar')
        
        pickle_name = utils.picklefy_project_name(project_name)
        pickle_file = year + "_" + pickle_name + ".pickle"
        pickle_path = os.path.join(pickle_jar, pickle_file)
        
        if self.data_platform == 's3':
            pickle_path = Path(pickle_path).as_posix().replace('s3:/', 's3://')
            fs = s3fs.S3FileSystem()
            open = fs.open
        else:
            open = builtins.open

        with open(pickle_path, 'rb') as f:
            project = pickle.load(f)

        # Check if project version matches session version & re-process if not
        try:
            if project.__version__ != self.__version__:
                project = self.process_project(project_name)
        except:
            project = self.process_project(project_name)

        # Remove plotter object if show_plots=False
        if not show_plots and project.plotter:
            project.plotter.close_plots()

        # Update filepaths to work for any user
        project.dashboard_dir = self.dashboard_dir
        project = self.__update_project_filepaths(project)

        # Update data cutoff date if not explicitly supplied
        if data_cutoff_date:
            try:
                self.data_cutoff_date = parser.parse(data_cutoff_date)
            except:
                self.data_cutoff_date = datetime.today()
        else:
            self.data_cutoff_date = datetime.today()

        # Add project to project_list
        self.project_list[project_name] = project

        # Load neighbor pickles if possible
        # `config_neighbor_sensors` is a set so we'll grab a copy of that & add in `neighbor_sensors`
        all_neighbors = project.config_neighbor_sensors.copy()
        all_neighbors.update(project.neighbor_list)
        for neighbor in all_neighbors:
            if neighbor not in self.project_list:
                self.load_pickle(neighbor)

        return project


    def load_pickles(self, project_list, year=None, data_cutoff_date=None, show_plots=False):
        """Calls [load_pickle()][ccrenew.dashboard.session.DashboardSession.load_pickle]
        for each [Project][ccrenew.dashboard.project.Project] in a list.

        Args:
            project_list (list): List of project names to load.
            year (int): See [load_pickle()][ccrenew.dashboard.session.DashboardSession.load_pickle]
            data_cutoff_date (str or datetime): See [load_pickle()][ccrenew.dashboard.session.DashboardSession.load_pickle]
            show_plots (bool): See [load_pickle()][ccrenew.dashboard.session.DashboardSession.load_pickle]
        """
        if isinstance(project_list, str):
            project_name = project_list
            self.load_pickle(project_name, year=year, data_cutoff_date=data_cutoff_date, show_plots=show_plots)
            return
        elif isinstance(project_list, list):
            for project_name in project_list:
                try:
                    project = self.load_pickle(project_name, year=year, data_cutoff_date=data_cutoff_date, show_plots=show_plots)
                    if project:
                        project.plotter.plot_list = {}
                except Exception as e:
                    print(e)
                    continue
        else:
            raise TypeError("`project_name` must be a string or Python list! Please reformat input and try again.")


    def draw_plots(self, project_name:str|Project|list, plot_order=[], reprocess:bool = False,
                        datasub:bool = False, use_solcast:bool = True, *args, **kwargs):
        """Draws plots for the given project.

        Args:
            project_name (str, Project, or list): A project name or list of project names
                to draw plots on for analysis. If a list is supplied it will plot
                the projects one-by-one. The user must press a button in the final
                plot of each project to move on to the next one.
            plot_order (list or str, optional): A list of plots to draw. This can
                be any number of plots from one to all. Defaults to None, which will draw the below plots in the order listed.
            reprocess (bool): Whether to reprocess the data. This option will likely
                be uncommon as any changes to a [Project's][ccrenew.dashboard.project.Project] config file or powertrack file will
                automatically reprocess that data. This could be used if any changes are made
                outside of the config/powertrack file or a change is made to a neighbor.
            datasub (bool): Option to use automatic data substitution algorithm.
            use_solcast (bool): Option to use Solcast data in data sub process. `datasub`
                must be set to `True` for this option to take effect.

        Default Plot Order

        | Plot Alias                | Plot Description                                          |
        |---------------------------|-----------------------------------------------------------|
        |    ``xplot_pwr_poa``      |    Crossplot of power meter & POA data                    |
        |    ``xplot_temp``         |    Crossplot of power meter & POA data, colored by Tamb   |
        |    ``temps``              |    Tcell temperature comparison                           |
        |    ``inv``                |    Inverters                                              |
        |    ``pr``                 |    Hourly Performance Ratio                               |
        |    ``8760``               |    Meter Corrected vs 8760                                |
        |    ``weather``            |    Weather sensors                                        |
        |    ``mtr_corrected``      |    Meter corrections                                      |
        |    ``mtr_dif``            |    Meter correction dif                                   |
        |    ``poas``               |    POA sensors                                            |
        |    ``pwr_poa``            |    Power meter & POA Avg timeseries plot                  |
        |    ``ghi``                |    GHI sensors                                            |
        |    ``irrad``              |    POA & GHI sensors                                      |
        |    ``tz``                 |    Timezone check                                         |
        |    ``losses``             |    Losses  by type                                        |
        |    ``poa_corr``           |    POA correlation check                                  |

        Other Optional Plots

        | Plot Alias                | Plot Description                                          |
        |---------------------------|-----------------------------------------------------------|
        |    ``meters``             |    Original & corrected power and cumulative meters       |
        |    ``native``             |    Native POA/GHI/Tamb/Tmod/Wind sensor subplots          |

        Keyword Args:
            mth (int): The month to show on the Power vs POA crossplot. Defaults to current month.
            default_tool (str): The default tool for navigating around the plots. Options
                are `zoom` and `pan`. Defaults to `zoom`.
            redraw (bool): Option to force the session to redraw the plots.
            open_folder (bool): Option to open the project's folder on the filesystem. Defaults to False.
            close_plots (bool, str, or list): Option to close all or selected open plots.
                If set to True all plots will be closed. If a plot alias or list of plot
                aliases are provided only those plots will be closed. Defaults to True.
            min_date (int, str, or datetime): Option to set a minimum date on the timeseries
                plots. If an integer is supplied the minimum date will be se to that
                many days previous to the current date. If a datetime or string representation
                of a datetime is supplied it will set the minimum date to that date. Defaults to None.
            fullscreen (bool): Option to show plots as fullscreen. Defaults to True.
            screen (int): 1-based index of screen to draw plots on. I.e. a workstation
                with 3 monitors would accept 1, 2, or 3 as an argument.
            poa_onboarding (bool): Plots POA correlation for all months instead of the
                selected month. Defaults to False.
        """
        if isinstance(project_name, (str, Project)):
            redraw = kwargs.get('redraw', False)
            project = self.process_project(project_name, reprocess=reprocess, datasub=datasub, use_solcast=use_solcast)
            neighbor_sensor_data = {self.project_list[neighbor].project_name: self.project_list[neighbor].df_sensors_native_avg for neighbor in project.neighbor_list}
            try:
                if project.errored:
                    raise Exception(f'{project.project_name} encountered an error while processing. Plots cannot be drawn at the moment. To investigate information on the error call `project.error_info`')
            except AttributeError:
                project.errored = False

            try:
                plotter = self.plotters[project_name]
            except KeyError:
                plotter = Plotter(project, neighbor_sensor_data)
                self.plotters[project_name] = plotter

            plotter._update_plotter_neighbors(neighbor_sensor_data)

            if not redraw:
                # Check if the project has an active plotter object & if the plot order is the same
                if plotter.plot_list == plot_order:
                    print('Plots already drawn. Add a new plot to the `plot_list` attribute or set `redraw` = True if you\'d like to redraw the current plots')
                else:
                    plotter.draw_plots(plot_order=plot_order, *args, **kwargs)
            else:
                plotter.draw_plots(plot_order=plot_order, *args, **kwargs)

        elif isinstance(project_name, list):
            project_list = project_name
            project_count = len(project_list)
            for i, project_nm in enumerate(project_list):
                project = self.get_project(project_nm, get_neighbors=False)
                self.draw_plots(project.project_name, *args, **kwargs)
                try:
                    print('Press any key in the last plot window to show plots for {}'.format(project_list[i+1]))
                except IndexError:
                    print('Last plot in the list')
                while True:
                    if i+1 == project_count:
                        break
                    if waitforbuttonpress():
                        break


                
    def __update_project_filepaths(self, project):
        """Sets filepaths for config file, Powertrack file, and Pickle Jar

        Args:
            project (Project): A `ccrenew.dashboard.project.Project`

        Returns:
            Project: a `ccrenew.dashboard.project.Project` with updated filepaths
        """
        # Build filename for config file
        project.config_filepath = project._find_config_file()     

        # Build filename for bool file
        project.bool_filepath = project._find_bool_file()     

        # Build filename for Powertrack file
        project.powertrack_filepath, project.powertrack_csv = project._find_powertrack_file()

        # Build filename for Powertrack csv

        # Check for pickle jar folder & create if it doesn't exist
        project.pickle_jar = project._find_pickle_jar() 
        
        return project
    

    def __prepare_source_data(self, project):

        # Make sure filepath references are correct for config, bool, & Powertrack files + Pickle Jar
        self.__update_project_filepaths(project)

        # If the config file hasn't been updated it will use the data that's already been pulled
        # If the config file has been updated it will read the file & update the data
        self.__update_project_config(project)

        # If the bool file hasn't been updated it will use the data that's already been pulled
        # If the bool file has been updated it will read the file & update the data
        project._load_bool_file()

        # If the powertrack file hasn't been updated it will use the data that's already been pulled
        # If the powertrack file has been updated it will read the file & update the data
        project.load_production_data()

        # fetch_solcast_data()
    

    def __update_project_config(self, project):
        
        # Store the list of sensors before pulling the config file
        sensors = project.Get_Sensor
        project._parse_config_file()

        # Check if new neighbor sensors have been added to the config & pull them if so
        if not sensors.equals(project.Get_Sensor):
            project._find_neighbor_sensors()
            self.__get_neighbor_sensor_data(project)
            
            
    def __get_neighbor_sensor_data(self, project):
        for neighbor_name in project.config_neighbor_sensors:
            # Update production data from neighbor
            try:
                neighbor = self.get_project(neighbor_name, get_neighbors=False)
            except KeyError:
                print(f'Neighbor for {project.project_name} not found: {neighbor_name}. Sensor data for the neighbor will be blank. Check logs to determine why neighbor was not added to DashboardSession')
                continue
            except Exception as e:
                print(f'Neighbor for {project.project_name} errored: {neighbor_name}. Sensor data for the neighbor will be blank. Check logs to determine why neighbor was not added to DashboardSession.\nError details: {e}')
                continue
            
            # Reload neighbor's config or powertrack files if needed
            self.__prepare_source_data(neighbor)

            # Find the columns needed from the neighbor
            sensor_cols = project.Get_Sensor.loc[project.Get_Sensor['Source'] == neighbor_name, 'Value'].tolist()

            # Get the columns from the neighbor
            try:
                neighbor_df = neighbor.df.reindex(index=project.df.index, columns=sensor_cols)
            # If the production data hasn't been loaded for the project
            except AttributeError:
                neighbor.load_production_data()
                neighbor_df = neighbor.df.reindex(index=project.df.index, columns=sensor_cols)
            # If the neighbor errored out during initialization we'll just create a blank df for the neighbor
            except NameError:
                neighbor_cols = [col + '_' + neighbor_name for col in sensor_cols]
                project.df = project.df.reindex(columns = project.df.columns.tolist() + neighbor_cols)
                project.df_sensor_ON = project.df_sensor_ON.reindex(columns = project.df_sensor_ON.columns.tolist() + neighbor_cols)
                return

            neighbor_df.rename(columns=lambda x: str(x) + '_' + neighbor_name, inplace=True)
            neighbor_sensor_ON = neighbor.df_sensor_ON.reindex(index=project.df_sensor_ON.index, columns=sensor_cols)
            neighbor_sensor_ON.columns = neighbor_df.columns.tolist()

            # Add neighbor columns from config file
            project.df = df_update_join(project.df, neighbor_df)
            project.df_sensor_ON = df_update_join(project.df_sensor_ON, neighbor_sensor_ON)
            project.df_sensor_ON = project.df_sensor_ON.reindex(columns=project.df.columns.tolist())

            # Add the neighbor columns to the positional references
            project._locate_column_positions()



# -*- coding: utf-8 -*-

"""
Created on Fri May 26 11:43:57 2017

-v3: adding conditions around interp. If outside the mesh, return nan and use nearest reporting site data

@author: BLUMENTHAL
"""

import pandas as pd
import numpy as np
import scipy.interpolate
import datetime
import math
import s3fs
import shapely.geometry as gs

from ccrenew.dashboard.utils.project_neighbors import haversine


def timeseries(raw_snow_df, project_df_index, project_lat, project_lon):

    X, Y = np.meshgrid([project_lon],[project_lat])
    df = raw_snow_df.copy()
    
    #Z = []
    
    #filter sites based on distance to project
    df['dist'] = ((df.Longitude - project_lon).pow(2) + (df.Latitude - project_lat).pow(2)).pow(0.5) * 69
    # df['dist'] = haversine(project_lon, project_lat, df['Longitude'], df['Latitude'])
    df = df.loc[ df['dist'] < 200, : ]
    
    #filter unused columns via production data index    
    # date_cols = [x for x in orig_df.columns if (type(x) in [datetime.datetime, pd.Timestamp])]
    date_cols = []
    for col in raw_snow_df.columns:
        try:
            pd.to_datetime(col)
            date_cols.append(col)
        except:
            continue
    df_index_dates = pd.to_datetime(list(set([x.date() for x in project_df_index])))

    
    #loop through applicable snow data columns
    match_dates = [col for col in date_cols if col in df_index_dates]
    
    #z_df = pd.DataFrame(index = match_dates, columns = ['snow']) 
    #create a snow df that is empty, to fill later
    z_df = pd.DataFrame(index = project_df_index, columns = ['snow'])
    z_df['snow'] = 0

    for col in match_dates:
        #create a df of the snow data, but just for each date
        t_df = df.copy().loc[~(df[col].isnull()), [col] + ['Longitude', 'Latitude', 'dist']]
        
        #if there is no snow file for the date, set to 0
        if t_df.size == 0:
            z_df.loc[col, 'snow'] = 0
            continue
        
        #requires 4 points within the 200 limit to interp. if less than 4 just get average of available points
        if len(t_df) < 4:
            z_df.loc[col,'snow'] = t_df[col].mean()
            continue
        
        z = t_df[col]
        x = t_df['Longitude']   #switched x and y 5/30
        y = t_df['Latitude']
       
        cartcoord = list(zip(x, y))
        
        #before interpolating, check to see if point is within mesh. If not then use closest point
        hull = gs.MultiPoint(cartcoord).convex_hull
        if hull.contains(gs.Point([project_lon, project_lat])):
            #print('contains')
            interp = scipy.interpolate.LinearNDInterpolator(cartcoord, z, fill_value= 0)
            #print(interp)
            Z0 = interp(X, Y)[0][0]
        else: 
            #Z0 = t_df.loc[t_df.dist.idxmin(), col] ##sometimes you get an array or a value?
            Z0 = np.median(t_df.loc[t_df.dist.idxmin(), col])
                
        #Z.append(Z0[0][0])
        z_df.loc[col, 'snow'] = Z0
        z_df[z_df['snow']<.01]=0   #added 12/1/2022 due to extraneous data with record low snow amounts in dataset. if less than 1/100th of 1 inch, not considered snow

    #df = pd.DataFrame(Z)
    #df.index = pd.date_range('2017-01-01', '2017-01-31')
    return z_df

#helper function to get data from s3
def retrieve_df(bucket, key):
    path = "s3://{b}/{k}".format(b=bucket, k=key)
    try:
        df = pd.read_csv(path, index_col=0, parse_dates=True,header=None)
        df = df.loc[~df.index.duplicated(), :]
    except IOError:
        return pd.DataFrame()
    return df

def coverage_v3(df, snow_data, aux,data_source,project_name):
    #take v2, but make the jan first value the same as the last value from previous year, if it exists
    
    bucket = 'perfdatadev.ccrenew.com'
    bucket_prefix_snow = "snow_projects/"

    POA = df['POA_avg'].copy()
    Tamb = df['Tamb_avg'].copy()
    
    #swap if necessary
    Tamb.loc[Tamb == 0] = aux
    
    coverage = pd.Series(index = df.index)
    coverage.iloc[0] = 0
    
    #try and get the last point of last year instead of defaulting at 0
    year = data_source.split('_')[0]
    if year=="AE":
        year = datetime.date.today().year
    year=int(year)
    last_year=year-1

    fname_data='{l}_data_snow_coverage_project.csv'.format(l=last_year)
    snow_key=bucket_prefix_snow+project_name+"/"+fname_data


    cover_last=retrieve_df(bucket, snow_key)
    if len(cover_last)>0:
        try:
            coverage.loc['{}-01-01 00:00:00'.format(year)] #fail if the first isnt in the new year
            #print('at least the first exists')
            last_value=cover_last.loc['{}-12-31 23:00:00'.format(last_year)].item()
            coverage.loc['{}-01-01 00:00:00'.format(year)]=last_value #this should do nothing if it doesnt start at 1/1
        except:
            print("Last Year's Snow file does not got to end of year or this year doesn't start on 1/1")
    
    m = -80
    theta = 20 * 3.14159/180
    for dt in df.index[0:-1]:
        # We'll get the date portion of the datetime to use for parsing the data
        dt_date = dt.floor('d')
        if dt_date in snow_data.index:                                #prevents issues with first day not starting at midnight
            snow = snow_data.loc[dt_date].item()
            snow_percent = np.min([snow / .01,1])
            if dt.hour == 0 and (snow_data.loc[dt_date] > 0).item():
                if snow_percent > coverage[dt]:
                    coverage[dt] = snow_percent
            #this is the section that allows for melting. if Tamb is below 0 continuously, it never melts and uses previous coverage value
            if Tamb[dt] > POA[dt] / m:
                coverage[dt] -= 0.197 * math.sin(theta)
            coverage[dt + pd.Timedelta("1h")] = coverage[dt]
    
    coverage[coverage < 0] = 0
    coverage.fillna(0)
    return coverage




from __future__ import annotations
import awswrangler as wr
import boto3
from dateutil import parser
import pandas as pd
from pathlib import Path
from multiprocessing.pool import ThreadPool
import os
import s3fs
import time

from ccrenew import (
    all_df_keys as df_keys,
    ccr,
    DateLike,
    Numeric,
    ProjectModelParams,
    AWSError,
)

s3_client = boto3.client('s3')
athena_client = boto3.client("athena")

RAW_DATA_BUCKET = 'perfdatadev.ccrenew.com'
BUCKET_PREFIX_SATELLITE = "5min_archive/Solcats/"
BUCKET_PREFIX_MEASURED = "5min_archive/PF/"
ATHENA_STAGING_DIR = "s3://cypress-perfeng-datalake-onprem-us-west-2/Temp/Athena_Results/"

# https://preview-docs.solcast.com.au/

def query_athena(query:str, database:str, params:list):
    query_response = athena_client.start_query_execution(
        QueryString=query,
        QueryExecutionContext={"Database": database},
        ExecutionParameters=params,
        ResultConfiguration={
            "OutputLocation": ATHENA_STAGING_DIR,
            "EncryptionConfiguration": {"EncryptionOption": "SSE_S3"},
        },
    )
    while True:
        try:
            # This function only loads the first 1000 rows
            athena_client.get_query_results(
                QueryExecutionId=query_response["QueryExecutionId"]
            )
            break
        except Exception as err:
            if "not yet finished" in str(err):
                time.sleep(0.001)
            else:
                raise err

    results_filename = query_response['QueryExecutionId'] + '.csv'
    results_filepath = ATHENA_STAGING_DIR+results_filename

    df = pd.read_csv(results_filepath)

    return df

def get_sat_weather(project_name:str, start:str|DateLike, end:str|DateLike,
                    convert:bool = False, pool_size:int = 6):
    """
    Query data from AWS to use in irradiance calculations
    
    Args:
        ccr_id (str): CCR ID for the project.
        start (str or DateLike): Start date for the data request.
        end (str or DateLike): Start date for the data request.
        source (str): Option to pull data from satellite recordings or on-site
            measurements. Options: `satellite` or `measured`
        pool_size (int): Number of threads to use when pulling data directly from S3.
    
    Returns
        df_weather (pd.DataFrame): Weather for the project over the specified days.
    """
    ccr_id, folder = _get_params(project_name)

    if isinstance(start, str):
        start = parser.parse(start)
    if isinstance(end, str):
        end = parser.parse(end)

    df_weather = query_athena(
        query="""SELECT  local_datetime
                        ,ghi
                        ,dni
                        ,dhi
                        ,air_temp
                        ,wind_speed_10m
                   FROM raw_data
                  WHERE ccr_id=?
                    AND local_datetime>=?
                    AND local_datetime<=?
               ORDER BY local_datetime;""",
        params=[f"'{ccr_id}'",
                f"CAST('{start.strftime('%Y-%m-%d')}' AS DATE)",
                f"CAST('{end.strftime('%Y-%m-%d')}' AS DATE)"
                ],
        database='solcast')

    # Drop duplicate dates - most common during DST change in the fall where we re-do 2 AM
    df_weather = df_weather.set_index('local_datetime') # type: ignore
    df_weather.index = pd.to_datetime(df_weather.index)
    df_weather = df_weather[~df_weather.index.duplicated()]

    if df_weather.empty:
        raise AWSError(f"Error querying weather data from AWS")
    else:
        df_weather = df_weather.rename(columns={'wind_speed_10m': 'Wind_speed',
                                                'air_temp': 'Tamb',
                                                'ghi': 'sat_ghi'})

        if convert:
            # Convert units
            folder = folder
            config_file = project_name + r'_Plant_Config_MACRO.xlsm'
            config_path = Path(ccr.file_project) / folder / project_name / 'Plant_Config_File' / config_file
            xl=pd.ExcelFile(config_path)
            df_config=xl.parse('Unit_Tab')
            
            #temp units
            temps = df_config.loc[df_config['Var_ID'].str.contains('amb')]['Convert_Farenheit_to_Celcius'].sum()
            temps += df_config.loc[df_config['Var_ID'].str.contains('mod')]['Convert_Farenheit_to_Celcius'].sum()
            if temps > 0:
                df_weather['Tamb'] = df_weather['Tamb']*float((9/5.))+32.0
                df_weather['Tmod'] = df_weather['Tmod']*float((9/5.))+32.0
            else:
                df_weather['Tamb'] = df_weather['air_temp']
                
            #wind units
            winds = df_config.loc[df_config['Var_ID'].str.contains('speed')]['Convert_mph_to_mps'].sum()
            if winds > 0:
                df_weather['Wind_speed'] = df_weather['wind_speed_10m']*float(2.23694)
            else:
                df_weather['Wind_speed'] = df_weather['wind_speed_10m']

    return df_weather

def get_project_ghi(project_name:str, start:str|DateLike, end:str|DateLike,
                    ghi_index:int = 0, pool_size:int = 6):
    ccr_id, _ = _get_params(project_name)
    keys, (start, end) = _generate_keys(ccr_id, start, end, 'project_ghi')
    df_project_ghi = _s3_pool(pool_size=pool_size, keys=keys)
    ghi_col = [col for col in df_project_ghi.columns if 'GHI' in col][ghi_index]
    df_project_ghi = df_project_ghi[[ghi_col]].rename(columns={ghi_col: 'proj_ghi'})

    return df_project_ghi

def _make_satellite_key(CCR_ID, date):
    return BUCKET_PREFIX_SATELLITE + CCR_ID + "/" + "sat_weather_" + date.strftime('%Y-%m-%d') + ".csv"

def _make_project_key(CCR_ID, date):
    return BUCKET_PREFIX_MEASURED + CCR_ID + "/" + "main_" + date.strftime('%Y-%m-%d') + ".csv"

def _retrieve_df(key):
    fs = s3fs.S3FileSystem(anon=False)
    path = "s3://{b}/{k}".format(b=RAW_DATA_BUCKET, k=key)
    try:
        with fs.open(path, 'rb') as f:
            df = pd.read_csv(f, index_col = 0, parse_dates = True)
            df = df.loc[~df.index.duplicated(), :]
    except FileNotFoundError:
        df = pd.DataFrame()
    return df

def _get_project_model_params(project_name:str) -> ProjectModelParams:

    df_proj_keys = df_keys.query("Project == @project_name").to_dict('records')[0]

    project_params = ProjectModelParams(
        project_name,
        df_proj_keys['CCR_ID'],
        df_proj_keys['Folder'],
        f"US/{df_proj_keys['Timezone']}",
        df_proj_keys['GPS_Lat'],
        df_proj_keys['GPS_Long'],
        df_proj_keys['Elevation'],
        df_proj_keys['MWDC'],
        df_proj_keys['Racking'],
        df_proj_keys['Tilt/GCR'],
        df_proj_keys['Max_angle'],
        df_proj_keys['Temp_Coeff_Pmax'],
        df_proj_keys['a_module'],
        df_proj_keys['b_module'],
        df_proj_keys['Delta_Tcnd'],
    )

    return project_params

def _get_params(project_name:str):
    project_params = df_keys.query("index == @project_name").to_dict(orient='records')[0]
    ccr_id = project_params['CCR_ID']
    folder = project_params['Folder']

    return ccr_id, folder

def _generate_keys(ccr_id:str, start:str|DateLike, end:str|DateLike, type: str):
    if isinstance(start, str):
        start = parser.parse(start)
    if isinstance(end, str):
        end = parser.parse(end)
    date_range = pd.date_range(start, end, normalize=True)

    if type == 'satellite':
        keys = [_make_satellite_key(ccr_id, date) for date in date_range]
    else:
        keys = [_make_project_key(ccr_id, date) for date in date_range]

    return keys, (start, end)

def _s3_pool(pool_size, keys):
    pool_size = 6
    pool = ThreadPool(pool_size)
    day_df_list = pool.map(_retrieve_df, keys)
    pool.close()
    pool.join()
    
    df_sat_weather = pd.concat(day_df_list, axis=0)
    return df_sat_weather




# -*- coding: utf-8 -*-

"""
Created on Wed Jun 22 14:16:25 2016

@author: SEBASTIAN

Change log:

- 2016/07/24 -> Added code to make sure that days 1/1 , 7/4 and 12/25 are holidays.  
- 2017/07/01 -> Fixed DST not working. DST start and end were read from config instead of being calculated
- 2017/07/13 -> Speed up code to run on peak hours for energy and capacity. 
- 2017/08/09 -> Add line-loss adjustment
TODO: read holiday list from config file

"""

import calendar
import datetime as dt
from holidays import UnitedStates
import numpy as np
import os
import pandas as pd
from pathlib import Path


from pandas.tseries.holiday import (
    AbstractHolidayCalendar,
    Holiday,
    nearest_workday,
    USMartinLutherKingJr,
    USPresidentsDay,
    GoodFriday,
    USMemorialDay,
    USLaborDay,
    USThanksgivingDay,
    TH,
    FR,
    SA,
    next_monday
)

class USTradingCalendar(AbstractHolidayCalendar):
    rules = [
        Holiday('NewYearsDay', month=1, day=1, observance=nearest_workday),
        #USMartinLutherKingJr,
        #USPresidentsDay,
        GoodFriday,
        USMemorialDay,
        Holiday('USIndependenceDay', month=7, day=4, observance=nearest_workday),
        USLaborDay,
        USThanksgivingDay,      
        Holiday('Christmas', month=12, day=25, observance=nearest_workday)
    ]

def unpivot(frame):
    frame_cols = frame.columns.tolist()
    output = pd.melt(frame, id_vars = frame_cols[0], value_vars = frame_cols[1:])
    return output

def get_trading_close_holidays(year):
    inst = USTradingCalendar()
    return inst.holidays(dt.datetime(year-1, 12, 31), dt.datetime(year, 12, 31))

def generate_Holidays(year):
    dr = pd.date_range(dt.datetime(year, 1, 1), dt.datetime(year+1, 1, 1), freq='H')    
    df = pd.DataFrame()
    df['Date'] = dr
    holidays = get_trading_close_holidays(year)
    # add day after Thanksgiving    
    Black_Friday = holidays[5] + dt.timedelta(days=1)    
    l_holidays = []
    #
    for i in range(0,len(holidays)):
        for j in range(0,24):
            l_holidays.append(holidays[i]  + dt.timedelta(hours = j) )
    #  add Black_Friday
    #
    for j in range(0,24):
            l_holidays.append(Black_Friday  + dt.timedelta(hours = j) )        
            
    df['Holiday'] = df['Date'].isin(l_holidays)    
    
    #df.set_value(idx, 'Holiday', True)
    df = df.set_index('Date')
	#
    #  Make sure that these days are holidays
    df.loc[(df.index.month == 1) & (df.index.day == 1), 'Holiday'] = True
    df.loc[(df.index.month == 7) & (df.index.day == 4), 'Holiday'] = True
    df.loc[(df.index.month == 12) & (df.index.day == 25), 'Holiday'] = True           

    return df

def yearly8760(df, project_name, year, shift_DST):
    
    #initialize dataframe
    df.rename(columns = {np.str(year): 'Rates' }, inplace= True)
    try:
        Flat_Capacity = df.loc['Flat_Capacity',:].values[0]
        df.drop('Flat_Capacity', inplace = True)
    except:
        Flat_Capacity = 0
    Flat_Total = Flat_Capacity + 0      #sum all yearly flat charges here: capacity, fees, etc
    
    #fix time index
    dr = pd.date_range(dt.datetime(year, 1, 1), dt.datetime(year+1, 1, 1), freq='H')
    if calendar.isleap(year):
        dr = dr[~((dr.month == 2) & (dr.day == 29))]
    if ((len(dr) == 8761) & (len(df) == 8760)):
        df.loc['add last_row',:] = 0
        df.set_index(dr, inplace = True)
    else:
        'something wrong with yearly rate calc, give up'
        print(project_name + ' Yearly Rate Calc Failed***')
        df = pd.DataFrame(np.zeros(len(dr)), index=dr, columns = ['Rates'])
        
    #Reverse DST shift - True for Pvsyst, False for AE data. Rate calc is already DST shifted as is the AE data. Must reverse DST shift for pvsyst
    if shift_DST:
        #second sunday in March
        d = dt.datetime(year, 3, 1)
        RATE_dst_start = d+ dt.timedelta(days= 7 + 7 - d.weekday() - 1)   
        #first sunday in November
        x = dt.datetime(year,11,1)
        RATE_dst_end = x + dt.timedelta(days= 7 - x.weekday() - 1)
        
        aux = pd.DataFrame([], index = df.index)
        aux['DST'] = 0
        aux.loc[((aux.index > RATE_dst_start) & (aux.index < RATE_dst_end)),:] = 1
        aux_2 = df[aux['DST'] == 1].shift(-1).fillna(0)
        df.loc[aux_2.index,'Rates'] = aux_2['Rates']
        
    #Fabricate leap day for leap years
    if calendar.isleap(year):
        dlp = pd.date_range(dt.datetime(year, 2, 29), dt.datetime(year, 3, 1), freq='H')[:-1]
        
        #get all the prices for the same day of the week
        aux = df.loc[(df.index.month ==2) & (df.index.dayofweek == dlp.dayofweek[0])]
        aux['hour'] = aux.index.hour
        leap_rates = aux.groupby('hour').mean().values
        
        d_add = pd.DataFrame(data = leap_rates, index = dlp, columns = ['Rates'])
        
        df = df.append(d_add)
        df = df.sort_index(ascending = True)
    
    df['Rates'] = df['Rates'].astype(float)
    df['Flat'] = Flat_Total / len(df.index)
    filler = ['Holiday','month','DST','weekday','Peak_day','ON_Peak','Summer','Energy_Peak','Capacity_Peak']
    
    aux = pd.DataFrame(np.zeros([len(df),len(filler)]), index = df.index, columns = filler)
    
    export = pd.concat([aux,df], axis = 1)
    
    return export

def generate_Rate_column (project_name, file_input, shift_DST, year, data_platform):

    def generate_rate(df_t1, df_t2, normal_rate_flag, year, shift_DST):
        #  Create DataFrame with Columns of Interest
        #
        #   Read all variables defined on Table 1
        #
        # year =  df_t1[df_t1['Name'] == 'Current Year']['Value'].values[0]
        #fixed 6/27/17, reads raw data instead of calculating DST dates
        #    RATE_dst_start = df_t1[df_t1['Name'] == 'DST_start']['Value'].values[0]
        #    RATE_dst_end = df_t1[df_t1['Name'] == 'DST_end']['Value'].values[0]
        d = dt.datetime(year, 3, 1)
        RATE_dst_start = d + dt.timedelta(days= 7 + 7 - d.weekday() - 1)
        
        x = dt.datetime(year,11,1)
        RATE_dst_end = x + dt.timedelta(days= 7 - x.weekday() - 1)
        
        On_Peak_Sunday    =  df_t1[df_t1['Name'] == 'On_Peak_Sunday']['Value'].values[0]
        On_Peak_Monday    =  df_t1[df_t1['Name'] == 'On_Peak_Monday']['Value'].values[0]   
        On_Peak_Tuesday   =  df_t1[df_t1['Name'] == 'On_Peak_Tuesday']['Value'].values[0]
        On_Peak_Wednesday =  df_t1[df_t1['Name'] == 'On_Peak_Wednesday']['Value'].values[0]
        On_Peak_Thursday  =  df_t1[df_t1['Name'] == 'On_Peak_Thursday']['Value'].values[0]
        On_Peak_Friday    =  df_t1[df_t1['Name'] == 'On_Peak_Friday']['Value'].values[0]
        On_Peak_Saturday  =  df_t1[df_t1['Name'] == 'On_Peak_Saturday']['Value'].values[0]
        
        Cap_Summer_1  =  df_t1[df_t1['Name'] == 'Cap_Summer_1']['Value'].values[0]
        Cap_Summer_2  =  df_t1[df_t1['Name'] == 'Cap_Summer_2']['Value'].values[0]
        Cap_Summer_3  =  df_t1[df_t1['Name'] == 'Cap_Summer_3']['Value'].values[0]
        Cap_Summer_4  =  df_t1[df_t1['Name'] == 'Cap_Summer_4']['Value'].values[0]
        Cap_Summer_5  =  df_t1[df_t1['Name'] == 'Cap_Summer_5']['Value'].values[0]
        Cap_Summer_6  =  df_t1[df_t1['Name'] == 'Cap_Summer_6']['Value'].values[0]
        Cap_Summer_7  =  df_t1[df_t1['Name'] == 'Cap_Summer_7']['Value'].values[0]
        Cap_Summer_8  =  df_t1[df_t1['Name'] == 'Cap_Summer_8']['Value'].values[0]
        Cap_Summer_9  =  df_t1[df_t1['Name'] == 'Cap_Summer_9']['Value'].values[0]
        Cap_Summer_10  =  df_t1[df_t1['Name'] == 'Cap_Summer_10']['Value'].values[0]
        Cap_Summer_11  =  df_t1[df_t1['Name'] == 'Cap_Summer_11']['Value'].values[0]
        Cap_Summer_12  =  df_t1[df_t1['Name'] == 'Cap_Summer_12']['Value'].values[0]
        
        LINE_LOSS = df_t1[df_t1['Name'] == 'Line Loss  Rate Increase (applies to ENERGY rates only)']['Value'].values[0]
            
        ENERGY_ON_PEAK  =  (1+LINE_LOSS) * df_t1[df_t1['Name'] == 'ENERGY_ON_PEAK_[$/KWh]']['Value'].values[0]
        ENERGY_OFF_PEAK =  (1+LINE_LOSS) * df_t1[df_t1['Name'] == 'ENERGY_OFF_PEAK_[$/KWh]']['Value'].values[0]
        
        CAPACITY_ON_PEAK_SUMMER =  df_t1[df_t1['Name'] == 'CAPACITY_ON_PEAK_SUMMER_[$/KWh]']['Value'].values[0]
        CAPACITY_ON_PEAK_NON_SUMMER =  df_t1[df_t1['Name'] == 'CAPACITY_ON_PEAK_NON_SUMMER_[$/KWh]']['Value'].values[0]
        
        
        #---------------
        #  Creates Index with Dates and Column with Holidays info
        df = generate_Holidays(year)
        #  
        df['month'] = df.index.month
        df['hour'] = df.index.hour
        #
        #---------------
        #  DST?
        #
        #    aux = pd.DataFrame(np.zeros(shape=(len(df), 1)), columns = ['DST'])
        #    aux = aux.set_index(df.index)
        #    aux_2 = aux[RATE_dst_start : RATE_dst_end] +1
        #    df['DST'] = aux.add(aux_2, fill_value = 0)
        df['DST'] = 0
        df.loc[(df.index > RATE_dst_start) & (df.index < RATE_dst_end), 'DST'] = 1
        
        #---------------
        #  Weekday
        df['weekday'] = df.index.weekday
        #
        #  Find holidays that are in Sundays
        #holidays_and_Sunday = df[df['weekday'] == 6 & df['holiday'] == ]
        #---------------
        #  Peak Day
        #  Monday = 0,..., Sunday = 6  (In excel is Sunday is 1)
        df['Peak_day'] = 0
        
        df.loc[df['weekday'] == 6, ['Peak_day']] = On_Peak_Sunday
        df.loc[df['weekday'] == 0, ['Peak_day']] = On_Peak_Monday
        df.loc[df['weekday'] == 1, ['Peak_day']] = On_Peak_Tuesday
        df.loc[df['weekday'] == 2, ['Peak_day']] = On_Peak_Wednesday
        df.loc[df['weekday'] == 3, ['Peak_day']] = On_Peak_Thursday
        df.loc[df['weekday'] == 4, ['Peak_day']] = On_Peak_Friday
        df.loc[df['weekday'] == 5, ['Peak_day']] = On_Peak_Saturday
        #
        df.loc[df['Holiday'] == True, ['Peak_day']] = 0
        
        #---------------
        #  On Peak  [AND( peak day, NOT Holiday)]
        
        df['ON_Peak'] = 0
        df.loc[(df['Peak_day'] == 1) & (df['Holiday'] == False), ['ON_Peak']] = 1
        
        #  Summer (1) or Not Summer (0)
        df['Summer'] = 0
        
        df.loc[df['month'] == 1, ['Summer']] = Cap_Summer_1
        df.loc[df['month'] == 2, ['Summer']] = Cap_Summer_2
        df.loc[df['month'] == 3, ['Summer']] = Cap_Summer_3
        df.loc[df['month'] == 4, ['Summer']] = Cap_Summer_4
        df.loc[df['month'] == 5, ['Summer']] = Cap_Summer_5
        df.loc[df['month'] == 6, ['Summer']] = Cap_Summer_6
        df.loc[df['month'] == 7, ['Summer']] = Cap_Summer_7
        df.loc[df['month'] == 8, ['Summer']] = Cap_Summer_8
        df.loc[df['month'] == 9, ['Summer']] = Cap_Summer_9
        df.loc[df['month'] == 10, ['Summer']] = Cap_Summer_10
        df.loc[df['month'] == 11, ['Summer']] = Cap_Summer_11
        df.loc[df['month'] == 12, ['Summer']] = Cap_Summer_12
        #    
        #---------------
        #  Energy Peak 
        #
        #  Extract table from Table_2 sheet.
        ep_t = df_t2.iloc[2:26].astype(int)
        # shift hours from 1-24 to 0-23
        ep_t['ENERGY PEAK HOURS'] -= 1
        #prep for unpivot and melt data
        ep_t.set_index(ep_t[ep_t.columns[0]], drop = True, inplace = True)
        ep_t.columns = range(0,13)
        frame = unpivot(ep_t)
        
        #commute peak hours back to df
        frame['string'] = frame[[0, u'variable']].astype(str).apply(lambda x: '--'.join(x), axis=1)
        lis = frame.loc[ frame['value'] == 1, 'string'].tolist()

        aux_df = pd.DataFrame([], df.index)
        aux_df['string'] = df[['hour', 'month']].astype(str).apply(lambda x: '--'.join(x), axis=1)
        
        df['Energy_Peak'] = 0
        df.loc[ aux_df['string'].isin(lis), 'Energy_Peak'] = 1
        
        aux_2 = []
        
            #  Multiply by Colun On Peak
        df['Energy_Peak'] = df['Energy_Peak'].multiply(df['ON_Peak'], axis = 'index')
        
        #  Shift Values in Energy Peak Based on DST Column
        
        if shift_DST == True:
            aux = df['Energy_Peak']    
            aux = aux [df['DST'] == 1].shift(-1).fillna(0)
            aux_2 =   pd.concat([df['Energy_Peak'] , aux])
            aux_2 = aux_2[~aux_2.index.duplicated(keep='last')]
            df['Energy_Peak'] = aux_2
        
        #---------------
        #  CAPACITY Peak? 
        #
        #  Extract table from Table_2 sheet.
        ep_t = df_t2.iloc[29:54].astype(int)
        # shift hours from 1-24 to 0-23
        ep_t['ENERGY PEAK HOURS'] -= 1
    
        ep_t.set_index(ep_t[ep_t.columns[0]], drop = True, inplace = True)
        ep_t.columns = range(0,13)
        frame = unpivot(ep_t)
        frame['string'] = frame[[0, u'variable']].astype(str).apply(lambda x: '--'.join(x), axis=1)
        lis = frame.loc[ frame['value'] == 1, 'string'].tolist()

        aux_df = pd.DataFrame([], df.index)
        aux_df['string'] = df[['hour', 'month']].astype(str).apply(lambda x: '--'.join(x), axis=1)
        
        df['Capacity_Peak'] = 0
        df.loc[ aux_df['string'].isin(lis), 'Capacity_Peak'] = 1

        
            #  Multiply by Colun On Peak
        df['Capacity_Peak'] = df['Capacity_Peak'].multiply(df['ON_Peak'], axis = 'index')
            #  Shift Values in Energy Peak Based on DST Column

        if shift_DST == True:
            aux = df['Capacity_Peak']    
            aux = aux [df['DST'] == 1].shift(-1).fillna(0)
            aux_2 =   pd.concat([df['Capacity_Peak'] , aux])
            aux_2 = aux_2[~aux_2.index.duplicated(keep='last')]
            df['Capacity_Peak'] = aux_2   
        #
        #---------------
        
        df['Rates'] = ENERGY_OFF_PEAK
        
        df.loc[(df['ON_Peak'] == 1) & (df['Summer'] == 1) & (df['Capacity_Peak'] == 1), ['Rates']] = ENERGY_ON_PEAK + CAPACITY_ON_PEAK_SUMMER 
        df.loc[(df['ON_Peak'] == 1) & (df['Summer'] == 0) & (df['Capacity_Peak'] == 1), ['Rates']] = ENERGY_ON_PEAK + CAPACITY_ON_PEAK_NON_SUMMER 
        
        df['Flat'] = 0              #unless read from monthly charges
        
        del df['hour']
        
        return df, normal_rate_flag


    try:
        df_t1 = pd.read_excel(file_input, sheet_name = 'Table_1')
        df_t2 = pd.read_excel(file_input, sheet_name = 'Table_2', index_col=0).fillna(0)
        normal_rate_flag = True
    except:
        #rate calc not in config file, redirect to yearly rates file
        normal_rate_flag = False

        #read new file - only use specific year's worth of data
        filedir = Path(file_input).parent
        filename = project_name + '_YearlyRates.csv'
        filepath = filedir / filename

        if data_platform == 's3':
            filepath = filepath.as_posix().replace('s3:/', 's3://')

        df = pd.read_csv(filepath, index_col = 'IND', usecols= ['IND', np.str(year)])

        if not shift_DST:
            df_1 = yearly8760(df.copy(), project_name, year, shift_DST=True)
            df_2 = yearly8760(df.copy(), project_name, year, shift_DST=False)

            return df_1, normal_rate_flag, df_2, normal_rate_flag
        else:
            df = yearly8760(df, project_name, year, shift_DST)

            return df, normal_rate_flag
    
    if not shift_DST:
        df1, rate_flag_1 = generate_rate(df_t1, df_t2, normal_rate_flag, year, shift_DST=True)
        df2, rate_flag_2 = generate_rate(df_t1, df_t2, normal_rate_flag, year, shift_DST=False)

        return df1, rate_flag_1, df2, rate_flag_2
    else:
        df, rate_flag = generate_rate(df_t1, df_t2, normal_rate_flag, year, shift_DST)

        return df, rate_flag